@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998, 1999  Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2002  Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,50 +16,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
-/*
- * $Log: CdrDriver.h,v $
- * Revision 1.14  1999/04/05 18:47:11  mueller
- * Added CD-TEXT support.
- *
- * Revision 1.13  1999/03/27 20:52:17  mueller
- * Added data track support for writing and toc analysis.
- *
- * Revision 1.12  1999/02/06 20:41:21  mueller
- * Added virtual member function 'checkToc()'.
- *
- * Revision 1.11  1998/10/24 14:28:59  mueller
- * Changed prototype of 'readDiskToc()'. It now accepts the name of the
- * audio file that should be placed into the generated toc-file.
- * Added virtual function 'readDisk()' that for reading disk toc and
- * ripping audio data at the same time.
- *
- * Revision 1.10  1998/10/03 15:10:38  mueller
- * Added function 'writeZeros()'.
- * Updated function 'writeData()'.
- *
- * Revision 1.9  1998/09/27 19:19:18  mueller
- * Added retrieval of control nibbles for track with 'analyzeTrack()'.
- * Added multi session mode.
- *
- * Revision 1.8  1998/09/22 19:15:13  mueller
- * Removed memory allocations during write process.
- *
- * Revision 1.7  1998/09/08 11:54:22  mueller
- * Extended disk info structure because CDD2000 does not support the
- * 'READ DISK INFO' command.
- *
- * Revision 1.6  1998/09/07 15:20:20  mueller
- * Reorganized read-toc related code.
- *
- * Revision 1.5  1998/08/30 19:12:19  mueller
- * Added supressing of error messages for some commands.
- * Added structure 'DriveInfo'.
- *
- * Revision 1.4  1998/08/25 19:24:07  mueller
- * Moved basic index extraction algorithm for read-toc to this class.
- * Added vendor codes.
- *
- */
 
 #ifndef __CDRDRIVER_H__
 #define __CDRDRIVER_H__
@@ -68,6 +24,7 @@
 #include "Msf.h"
 #include "TrackData.h"
 #include "SubChannel.h"
+#include "remote.h"
 
 class Toc;
 class Track;
@@ -75,6 +32,24 @@ class Track;
 #define OPT_DRV_GET_TOC_GENERIC   0x00010000
 #define OPT_DRV_SWAP_READ_SAMPLES 0x00020000
 #define OPT_DRV_NO_PREGAP_READ    0x00040000
+#define OPT_DRV_RAW_TOC_BCD       0x00080000
+#define OPT_DRV_RAW_TOC_HEX       0x00100000
+#define OPT_DRV_NO_CDTEXT_READ    0x00200000
+
+
+// reading capabilities
+#define CDR_READ_CAP_AUDIO_PW_RAW    0x001
+#define CDR_READ_CAP_AUDIO_RW_COOKED 0x002
+#define CDR_READ_CAP_AUDIO_RW_RAW    0x004
+#define CDR_READ_CAP_AUDIO_PQ_BCD    0x008
+#define CDR_READ_CAP_AUDIO_PQ_HEX    0x010
+#define CDR_READ_CAP_DATA_PW_RAW     0x020
+#define CDR_READ_CAP_DATA_RW_COOKED  0x040
+#define CDR_READ_CAP_DATA_RW_RAW     0x080
+#define CDR_READ_CAP_DATA_PQ_BCD     0x100
+#define CDR_READ_CAP_DATA_PQ_HEX     0x200
+
+#define CDR_AUDIO_SCAN_CAP (CDR_READ_CAP_AUDIO_PW_RAW|CDR_READ_CAP_AUDIO_PQ_BCD|CDR_READ_CAP_AUDIO_PQ_HEX)
 
 struct DiskInfo {
   long capacity;          // recordable capacity of medium
@@ -111,6 +86,19 @@ struct DriveInfo {
   int currentWriteSpeed;
 
   unsigned int accurateAudioStream : 1;
+  unsigned int burnProof : 1;
+  unsigned int ricohJustLink : 1;
+  unsigned int ricohJustSpeed : 1;
+};
+
+struct CdTextPack {
+  unsigned char packType;
+  unsigned char trackNumber;
+  unsigned char sequenceNumber;
+  unsigned char blockCharacter;
+  unsigned char data[12];
+  unsigned char crc0;
+  unsigned char crc1;
 };
 
 struct CdToc {
@@ -145,15 +133,6 @@ struct TrackInfo {
   long bytesWritten;    // number of bytes written to file
 };
 
-struct CdTextPack {
-  unsigned char packType;
-  unsigned char trackNumber;
-  unsigned char sequenceNumber;
-  unsigned char blockCharacter;
-  unsigned char data[12];
-  unsigned char crc0;
-  unsigned char crc1;
-};
 
 class CdrDriver {
 public:
@@ -163,6 +142,9 @@ public:
   // returns stored SCSI interface object
   virtual ScsiIf *scsiIf() const { return scsiIf_; }
 
+  // sets SCSI interface object
+  void scsiIf(ScsiIf *i) { scsiIf_ = i; }
+
   // returns name of driver
   virtual const char *driverName() const { return driverName_; }
 
@@ -173,6 +155,9 @@ public:
   // 0 for little endian byte order
   virtual int bigEndianSamples() const = 0;
 
+  // return information about drive
+  virtual const DriveInfo *driveInfo(int showErrorMsg) { return NULL; }
+
   // returns current writing speed
   virtual int speed() { return speed_; }
   
@@ -180,6 +165,23 @@ public:
   // this function may send SCSI commands to the drive
   virtual int speed(int) = 0;
 
+  // sets/return buffer under run protection setting (if supported by
+  // the drive: 1 = enabled, 0 = disbaled
+  virtual int bufferUnderRunProtection() const {
+    return enableBufferUnderRunProtection_;
+  }
+
+  virtual void bufferUnderRunProtection(int s) {
+    enableBufferUnderRunProtection_ = s != 0 ? 1 : 0;
+  }
+    
+  // sets/return writing speed control setting (if supported by
+  // the drive: 1 = enabled, 0 = disbaled
+  virtual int writeSpeedControl() const { return enableWriteSpeedControl_; }
+
+  virtual void writeSpeedControl(int s) {
+    enableWriteSpeedControl_ = s != 0 ? 1 : 0; }
+
   // returns 1 if simulation mode, 0 for real writing
   virtual int simulate() const { return simulate_; }
 
@@ -201,6 +203,9 @@ public:
   virtual int rawDataReading() const { return rawDataReading_; }
   virtual void rawDataReading(int f) { rawDataReading_ = f != 0 ? 1 : 0; }
 
+  virtual TrackData::SubChannelMode subChanReadMode() const { return subChanReadMode_; }
+  virtual void subChanReadMode(TrackData::SubChannelMode m) { subChanReadMode_ = m; }
+
   // Sets/returns the pad first pre-gap flag
   virtual int padFirstPregap() const { return padFirstPregap_; }
   virtual void padFirstPregap(int f) { padFirstPregap_ = f != 0 ? 1 : 0; }
@@ -212,9 +217,36 @@ public:
   // if 'fd' is >= 0 and clears it otherwise
   virtual void onTheFly(int fd);
 
+  // Returns force flag
+  virtual int force() const { return force_; }
+
+  // Sets force flag
+  virtual void force(int f) { force_ = f != 0 ? 1 : 0; }
+
+  // Returns TAO source flag
+  virtual int taoSource() const { return taoSource_; }
+
+  // Sets TAO source flag
+  virtual void taoSource(int f) { taoSource_ = f != 0 ? 1 : 0; }
+
+  // Return number of adjust sectors for reading TAO source disks
+  virtual int taoSourceAdjust() const { return taoSourceAdjust_; }
+
+  // Sets number of adjust sectors for reading TAO source disks
+  virtual void taoSourceAdjust(int val);
+
+  // Sets remote mode
+  virtual void remote(int flag, int fd);
+
+  // Return remote mode flag
+  virtual int remote() { return remote_; }
+  
   // Sets cdda paranoia mode
   void paranoiaMode(int);
 
+  // Return byte order of host (0: little endian, 1: big endian)
+  int hostByteOrder() const { return hostByteOrder_; }
+
   // general commands
   virtual int testUnitReady(int) const;
 
@@ -230,12 +262,23 @@ public:
 
   virtual int readCapacity(long *length, int showMessage = 1);
 
-  // CD-RW specific commands
+  virtual bool readBufferCapacity(long* total, long* available) {
+    return false;
+  }
 
-  virtual int blankDisk();
+  // CD-RW specific commands
+  
+  enum BlankingMode { BLANK_FULL, BLANK_MINIMAL };
+  virtual int blankDisk(BlankingMode);
 
   // disk at once recording related commands
 
+  // Returns acceptable sub-channel encoding mode for given sub-channel type:
+  //  -1: writing of sub-channel type not supported at all
+  //   0: accepts plain data without encoding
+  //   1: accepts only completely encoded data
+  virtual int subChannelEncodingMode(TrackData::SubChannelMode) const;
+
   // Should check if toc is suitable for DAO writing with the actual driver.
   // Returns 0 if toc is OK, else 1.
   // Usually all tocs are suitable for writing so that the base class
@@ -261,7 +304,8 @@ public:
 
   // Sends given data to drive. 'lba' should be the current writing address
   // and will be updated according to the written number of blocks.
-  virtual int writeData(TrackData::Mode, long &lba, const char *buf, long len);
+  virtual int writeData(TrackData::Mode, TrackData::SubChannelMode sm,
+			long &lba, const char *buf, long len);
 
   // returns mode for main channel data encoding, the value is used by
   // Track::readData()
@@ -285,7 +329,17 @@ public:
 
   // Returns block size depending on given sector mode and 'encodingMode_'
   // that must be used to send data to the recorder.
-  virtual long blockSize(TrackData::Mode) const;
+  virtual long blockSize(TrackData::Mode, TrackData::SubChannelMode) const;
+
+  // sends a status message to the driving application if in remote mode
+  enum WriteCdProgressType { WCD_LEADIN = PGSMSG_WCD_LEADIN,
+			     WCD_DATA = PGSMSG_WCD_DATA,
+			     WCD_LEADOUT = PGSMSG_WCD_LEADOUT };
+  int sendWriteCdProgressMsg(WriteCdProgressType type, int totalTracks,
+			     int track, int trackProgress, int totalProgress,
+			     int bufferFillRate, int writeBufferFill = 0);
+
+  int sendBlankCdProgressMsg(int totalProgress);
 
 
   // static functions
@@ -309,30 +363,46 @@ public:
   static int cdrVendor(Msf &, const char **vendor, const char** mediumType);
 
 protected:
+  struct ReadDiskInfo {
+    int tracks;    // total number of tracks
+    long startLba;      // LBA where extraction starts
+    long endLba;        // LBA where extraction ends
+  };
 
   unsigned long options_; // driver option flags
   ScsiIf *scsiIf_;
+  int scsiMaxDataLen_;
   char *driverName_;
 
   int hostByteOrder_; // 0: little endian, 1: big endian
 
+  unsigned long readCapabilities_;
+
   int blockLength_; // length of data block for 'writeData' command
   long blocksPerWrite_; // number of blocks that can be written with a
                         // single SCSI WRITE command
   char *zeroBuffer_; // zeroed buffer for writing zeros
-  
+
+  int enableBufferUnderRunProtection_;
+  int enableWriteSpeedControl_;
   int speed_;
   int simulate_;
   int multiSession_;
   int encodingMode_; // mode for encoding data sectors
   int fastTocReading_;
   int rawDataReading_;
+  TrackData::SubChannelMode subChanReadMode_;
   int padFirstPregap_; // used by 'read-toc': defines if the first audio 
                        // track's pre-gap is padded with zeros in the toc-file
                        // or if it is taken from the data file
   int onTheFly_; // 1 if operating in on-the-fly mode
   int onTheFlyFd_; // file descriptor for on the fly data
-
+  int force_; // force flag to allow certain operations
+  int remote_; // 1 for remote mode, else 0
+  int remoteFd_; // file descriptor for remote messages
+  int taoSource_; // 1 to indicate a TAO writting source CD for read-cd/read-toc
+  int taoSourceAdjust_; // number of unreadable sectors between two tracks
+                        // written in TAO mode
   const Toc *toc_;
 
   SubChannel **scannedSubChannels_;
@@ -345,6 +415,7 @@ protected:
   int audioDataByteOrder_; 
 
   static unsigned char syncPattern[12];
+  static char REMOTE_MSG_SYNC_[4];
 
   static int speed2Mult(int);
   static int mult2Speed(int);
@@ -369,7 +440,8 @@ protected:
 			   const unsigned char *modePageHeader,
 			   const unsigned char *blockDesc, int showErrorMsg);
 
-  virtual int writeZeros(TrackData::Mode, long &lba, long encLba, long count);
+  virtual int writeZeros(TrackData::Mode, TrackData::SubChannelMode,
+			 long &lba, long encLba, long count);
 
 
   // Returns track control flags for given track, bits 0-3 are always zero
@@ -415,6 +487,8 @@ protected:
   // a MODE2, MODE2_FORM1 or MODE2_FORM2 sector
   TrackData::Mode analyzeSubHeader(unsigned char *);
 
+  virtual unsigned long getReadCapabilities(const CdToc *, int) const = 0;
+
   // Called by 'readDiskToc()' to retrieve following information about
   // the track 'trackNr' with given start/end lba addresses:
   // - all index increments, filled into 'index'/'indexCnt'
@@ -466,7 +540,8 @@ protected:
   // Audio data that is usually retrieved with the sub-channels is placed
   // in 'buf' if it is not NULL.
   // Used by 'analyzeTrackScan()' and 'readAudioRangeParanoia()'.
-  virtual int readSubChannels(long lba, long len, SubChannel ***, Sample *buf);
+  virtual int readSubChannels(TrackData::SubChannelMode, long lba, long len,
+			      SubChannel ***, Sample *buf) = 0;
 
   // Determines the readable length of a data track and the pre-gap length
   // of the following track. The implementation in the base class should
@@ -479,21 +554,25 @@ protected:
   // the returned value will be smaller than 'len' down to 0. If a read
   // error occus -1 is returned. If a L-EC error occures -2 is returned.
   // This method is used by 'readDataTrack'/'analyzeDataTrack' and must be
-  // overloaded by the actual driver.
-  virtual long readTrackData(TrackData::Mode mode, long lba, long len,
-			     unsigned char *buf);
+  // overloaded by the driver.
+  virtual long readTrackData(TrackData::Mode, TrackData::SubChannelMode,
+			     long lba, long len, unsigned char *buf) = 0;
 
   // Reads a complete data track and saves data to a file.
-  virtual int readDataTrack(int fp, long start, long end,
+  virtual int readDataTrack(ReadDiskInfo *, int fp, long start, long end,
 			    TrackInfo *trackInfo);
 
   // Reads the audio data of given audio track range 'startTrack', 'endTrack'.
   // 'trackInfo' is am array of TrackInfo structures for all tracks. 
   // This function is called by 'readDisk()' and must be overloaded by the
   // actual driver.
-  virtual int readAudioRange(int fp, long start, long end,
+  virtual int readAudioRange(ReadDiskInfo *, int fp, long start, long end,
 			     int startTrack, int endTrack, 
-			     TrackInfo *trackInfo) = 0;
+			     TrackInfo *) = 0;
+
+  virtual int readAudioRangeStream(ReadDiskInfo *, int fd, long start,
+				   long end, int startTrack, int endTrack, 
+				   TrackInfo *);
 
   // Reads catalog number by scanning the sub-channels.
   // Uses 'readSubChannels()' to read the the sub-channels.
@@ -512,22 +591,34 @@ protected:
   Toc *buildToc(TrackInfo *trackInfos, long nofTrackInfos, int padFirstPregap);
 
   // sets block size for read/write operations
-  virtual int setBlockSize(long blocksize);
+  virtual int setBlockSize(long blocksize, unsigned char density = 0);
+
+  // checks if drive capabilities support requested sub-channel reading mode
+  int checkSubChanReadCaps(TrackData::Mode, unsigned long caps);
 
   void printCdToc(CdToc *toc, int tocLen);
 
+  enum ReadCdProgressType { RCD_ANALYZING = PGSMSG_RCD_ANALYZING,
+			    RCD_EXTRACTING = PGSMSG_RCD_EXTRACTING };
+  void sendReadCdProgressMsg(ReadCdProgressType, int totalTracks, int track,
+			     int trackProgress, int totalProgress);
+
 
-  // Interface for Monty's paranoia library:
 public:
-  // function called from 'cdda_read()' to read the audio data, currently
+  // function to read audio data and also the sub-channel data from
+  // specified lba,
+  // this function is called from 'cdda_read()', so that it is currently
   // public because I did not manage to define a friend function that has
   // C linkage :)
-  long paranoiaRead(Sample *buffer, long startLba, long len);
+  long audioRead(TrackData::SubChannelMode, int byteOrder,
+		 Sample *buffer, long startLba, long len);
 
+
+  // Interface for Monty's paranoia library:
 protected:
   // Extracts audio data for given track range with the help of 
   // Monty's paranoia library.
-  int readAudioRangeParanoia(int fp, long start, long end,
+  int readAudioRangeParanoia(ReadDiskInfo *, int fp, long start, long end,
 			     int startTrack, int endTrack, 
 			     TrackInfo *trackInfo);
 
@@ -536,16 +627,18 @@ private:
   void *paranoia_;                    // paranoia structure
   struct cdrom_drive *paranoiaDrive_; // paranoia device
   int paranoiaMode_;                  // paranoia mode
-  TrackInfo *paranoiaTrackInfo_;
-  int paranoiaStartTrack_;
-  int paranoiaEndTrack_;
-  long paranoiaActLba_;
-  int paranoiaActTrack_;
-  int paranoiaActIndex_;
-  long paranoiaCrcCount_;
-  int paranoiaError_;
-  long paranoiaProgress_;
-
+  ReadDiskInfo *audioReadInfo_;
+  TrackInfo *audioReadTrackInfo_;
+  int audioReadStartTrack_;
+  int audioReadEndTrack_;
+  long audioReadLastLba_;
+  long audioReadActLba_;
+  int audioReadActTrack_;
+  int audioReadActIndex_;
+  long audioReadCrcCount_;
+  int audioReadError_;
+  long audioReadProgress_;
+  
   // callback for the paranoia library, does nothing, currently
   static void paranoiaCallback(long, int);
 
