@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998  Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2002  Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,40 +16,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
-/*
- * $Log: PlextorReader.cc,v $
- * Revision 1.9  1999/04/05 18:48:37  mueller
- * Added driver options.
- * Added read-cd patch from Leon Woestenberg.
- *
- * Revision 1.8  1999/03/27 20:56:39  mueller
- * Changed toc analysis interface.
- * Added support for toc analysis of data disks.
- *
- * Revision 1.7  1998/09/27 19:20:05  mueller
- * Added retrieval of control nibbles for track with 'analyzeTrack()'.
- *
- * Revision 1.6  1998/09/08 11:54:22  mueller
- * Extended disk info structure because CDD2000 does not support the
- * 'READ DISK INFO' command.
- *
- * Revision 1.5  1998/09/07 15:20:20  mueller
- * Reorganized read-toc related code.
- *
- * Revision 1.4  1998/09/06 13:34:22  mueller
- * Use 'message()' for printing messages.
- *
- * Revision 1.3  1998/08/30 19:25:43  mueller
- * Added function 'diskInfo()'.
- * Changed sub channel data field in 'readSubChannelData()' to be more
- * compatible to other drives.
- *
- * Revision 1.2  1998/08/25 19:26:07  mueller
- * Moved basic index extraction algorithm to class 'CdrDriver'.
- *
- */
-
-static char rcsid[] = "$Id: PlextorReader.cc,v 1.9 1999/04/05 18:48:37 mueller Exp mueller $";
 
 #include <config.h>
 
@@ -68,12 +34,16 @@ static char rcsid[] = "$Id: PlextorReader.cc,v 1.9 1999/04/05 18:48:37 mueller E
 PlextorReader::PlextorReader(ScsiIf *scsiIf, unsigned long options)
   : CdrDriver(scsiIf, options)
 {
-  driverName_ = "Plextor CD-ROM Reader - Version 1.1";
+  driverName_ = "Plextor CD-ROM Reader - Version 1.3";
   
   speed_ = 0;
   simulate_ = 0;
   audioDataByteOrder_ = 0;
-
+  
+  slow_down_on_read_errors = -1;
+  transfer_data_before_max_speed = -1;
+  slow_down_on_vibrations = -1;
+  
   memset(&diskInfo_, 0, sizeof(DiskInfo));
   diskInfo_.valid.empty = 1;
 
@@ -105,12 +75,130 @@ PlextorReader::PlextorReader(ScsiIf *scsiIf, unsigned long options)
         m++;
     }
     model_ = models[m].number; // zero if not found
-
-    message(3, "model number %d\n",model_);
-    message(3, "PRODUCT ID: '%s'\n", scsiIf_->product());
+   
+	if (model_ > 0)	{
+    /* Plextor special features initialization */
+        unsigned char buf[32];
+        unsigned char header[8];
+        unsigned char blockdesc[8];
+        struct plex_msg {
+            int value;
+            char *msg;
+        };
+        struct plex_msg slowdown_msg[] = {
+            { -1, "Unsupported"},
+            { 0, "Enabled, unit will slow down"},
+            { 1, "Disabled, unit won't slow down"}
+        };
+        struct plex_msg trbefmax_msg[] = {
+            { -1, "Unsupported"},
+            { 0, "Disabled, unit will wait for max speed"},
+            { 1, "Enabled, unit will transfer as soon as possible"}
+        };
+        /* Read error slowdown seems available for all drives, while
+        vibration and transfer control only for PX-20 and later
+        Data is taken from
+        Mode page 0x31 byte 3 [ x x x x x sl td re ]
+        re=read errors , if 0 slow down on read errors, 0 default
+        td=transfer data, if 0 wait for max speed before transfer data, 0 default
+        sl=if 0 slowdown to avoid vibrations, 0 default
+        */
+        slow_down_on_read_errors = -1;
+        transfer_data_before_max_speed = -1;
+        slow_down_on_vibrations = -1;
+        orig_byte3 = 0;
+        if (getModePage6 (0x31, buf, 32, header, blockdesc, 1) == 0)    {
+            orig_byte3 = buf[3];
+            slow_down_on_read_errors = buf[3] & 0x01;
+            if (model_ >= 7)    {
+                transfer_data_before_max_speed = (buf[3] & 0x02) >> 1;
+                slow_down_on_vibrations = (buf[3] & 0x04) >> 2;
+            }
+        }
+        if (options & OPT_PLEX_NOSLOW_ON_ERR)
+            ReadErrorsSlowDown (0);
+        if (options & OPT_PLEX_TRANSF_BEF_MAX)
+            WaitMaxSpeed (0);
+        if (options & OPT_PLEX_NOSLOW_ON_VIB)
+            VibrationsSlowDown (0);
+        message (4, "Plextor features status:");
+        message (4, "  Slowdown on read errors: %s",
+            slowdown_msg[slow_down_on_read_errors+1].msg);
+        message (4, "  Slowdown to avoid vibrations: %s",
+            slowdown_msg[slow_down_on_vibrations+1].msg);
+        message (4, "  Transfer data before max speed: %s",
+            trbefmax_msg[transfer_data_before_max_speed+1].msg);
+    }
+    message(4, "model number %d\n",model_);
+    message(4, "PRODUCT ID: '%s'\n", scsiIf_->product());
   }
 }
 
+PlextorReader::~PlextorReader ()        {
+	/* Only for Plextor units */
+	if (model_ > 0)	{
+    	unsigned char buf[32];
+    	unsigned char header[8];
+    	unsigned char blockdesc[8];
+    	if (getModePage6 (0x31, buf, 32, header, blockdesc, 1) != 0) 
+        	return;
+    	buf[3] = orig_byte3;
+    	setModePage6 (buf, header, blockdesc, 1);
+	}
+}
+
+int PlextorReader::ReadErrorsSlowDown (int slowdown)    {
+    /* Not supported */
+    if (slow_down_on_read_errors == -1 ) return -1;
+    slowdown ? slowdown = 0 : slowdown = 1;
+
+    unsigned char buf[32];
+    unsigned char header[8];
+    unsigned char blockdesc[8];
+    if (getModePage6 (0x31, buf, 32, header, blockdesc, 1) != 0) 
+        return -1;
+    if (slowdown) 
+        buf[3] |= 0x01;
+    if (setModePage6 (buf, header, blockdesc, 1) != 0) 
+        return -1;
+    slow_down_on_read_errors = slowdown;
+    return 1;
+}
+
+int PlextorReader::VibrationsSlowDown (int slowdown)    {
+    if (slow_down_on_vibrations == -1) return -1;
+    slowdown ? slowdown = 0 : slowdown = 1;
+    
+    unsigned char buf[32];
+    unsigned char header[8];
+    unsigned char blockdesc[8];
+    if (getModePage6 (0x31, buf, 32, header, blockdesc, 1) != 0) 
+        return -1;
+    if (slowdown) 
+        buf[3] |= 0x04;
+    if (setModePage6 (buf, header, blockdesc, 1) != 0) 
+        return -1;
+    slow_down_on_vibrations = slowdown;
+    return 1;
+}
+
+int PlextorReader::WaitMaxSpeed (int wait)      {
+    if (transfer_data_before_max_speed == -1) return -1;
+    wait ? wait = 0 : wait = 1;
+    
+    unsigned char buf[32];
+    unsigned char header[8];
+    unsigned char blockdesc[8];
+    if (getModePage6 (0x31, buf, 32, header, blockdesc, 1) != 0) 
+        return -1;
+    if (wait) 
+        buf[3] |= 0x02;
+    if (setModePage6 (buf, header, blockdesc, 1) != 0) 
+        return -1;
+    transfer_data_before_max_speed = wait;
+    return 1;
+}
+
 // static constructor
 CdrDriver *PlextorReader::instance(ScsiIf *scsiIf, unsigned long options)
 {
@@ -363,7 +451,7 @@ CdRawToc *PlextorReader::getRawToc(int sessionNr, int *len)
 
   dataLen = ((reqData[0] << 8) | reqData[1]) + 2;
 
-  message(3, "Raw toc data len: %d", dataLen);
+  message(4, "Raw toc data len: %d", dataLen);
 
   data = new (unsigned char)[dataLen];
   
@@ -402,8 +490,9 @@ CdRawToc *PlextorReader::getRawToc(int sessionNr, int *len)
   return rawToc;
 }
 
-long PlextorReader::readTrackData(TrackData::Mode mode, long lba, long len,
-				  unsigned char *buf)
+long PlextorReader::readTrackData(TrackData::Mode mode,
+				  TrackData::SubChannelMode,
+				  long lba, long len, unsigned char *buf)
 {
   unsigned char cmd[10];
   long blockLen = 2340;
@@ -490,9 +579,9 @@ long PlextorReader::readTrackData(TrackData::Mode mode, long lba, long len,
 	   (actMode == TrackData::MODE2 ||
 	    actMode == TrackData::MODE2_FORM1 ||
 	    actMode == TrackData::MODE2_FORM2)))) {
-      message(3, "Stopped because sector with not matching mode %s found.",
+      message(4, "Stopped because sector with not matching mode %s found.",
 	      TrackData::mode2String(actMode));
-      message(3, "%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
+      message(4, "%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
 	      sector[0], sector[1], sector[2], sector[3],  sector[4],
 	      sector[5], sector[6], sector[7], sector[8], sector[9],
 	      sector[10], sector[11]);
@@ -559,7 +648,8 @@ Toc *PlextorReader::readDisk(int session, const char *dataFilename)
 // Creates 'Toc' object for inserted CD.
 // audioFilename: name of audio file that is placed into TOC
 // Return: newly allocated 'Toc' object or 'NULL' on error
-int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
+int PlextorReader::readAudioRangePlextor(ReadDiskInfo *rinfo, int fd,
+					 long startLba, long endLba,
 					 int startTrack, int endTrack, 
 					 TrackInfo *info)
 {
@@ -570,6 +660,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
  
   int blockLength = AUDIO_BLOCK_LEN + 1 + 294 + 96;
   long  blocksPerRead = scsiIf_->maxDataLen() / blockLength;
+  long stampLba;
 
   // first, last, current and number of recorded tracks
   int fat=-1,lat=-1,cat,nat=0;
@@ -612,7 +703,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
 
   cat = fat + 1;
  
-  message(1, "Reading CDDA blocks in range [%lu,%lu>"
+  message(2, "Reading CDDA blocks in range [%lu,%lu>"
     " (%lu blocks=%lu bytes).",fab,lab,nab,nab*2352);
 
   memset(cmd,0,12);
@@ -629,9 +720,12 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
   // set remaining audio blocks
   rab=nab;
 
+  stampLba = cab;
+
   // start optimistic
   speed(overspeed=20);
 
+
   // while still audio blocks left to read
   while (rab != 0)
   { 
@@ -651,6 +745,37 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
 
     message(1, "block %6ld\r",cab);
 
+    if (remote_) {
+      if (cab > stampLba + 75) {
+	long totalLen = info[cat + 1].start - info[cat].start;
+	long progress = cab - info[cat].start;
+	long totalProgress;
+
+	if (progress > 0) {
+	  progress *= 1000;
+	  progress /= totalLen;
+	}
+	else {
+	  progress = 0;
+	}
+
+	totalProgress = cab - rinfo->startLba;
+
+	if (totalProgress > 0) {
+	  totalProgress *= 1000;
+	  totalProgress /= rinfo->endLba - rinfo->startLba;
+	}
+	else {
+	  totalProgress = 0;
+	}
+
+	sendReadCdProgressMsg(RCD_EXTRACTING, rinfo->tracks, cat + 1, progress,
+			      totalProgress);
+
+	stampLba = cab;
+      }
+    }
+
     // SCSI command failed?
     while (sendCmd(cmd,12,NULL,0,data,n*blockLength))
     {
@@ -695,7 +820,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
 
           if ((chan.indexNr()!=cai) || (chan.trackNr()!=cat))
           {
-            message(1, "(track,index,absolute)=(%2d,%2d,%s)",
+            message(2, "(track,index,absolute)=(%2d,%2d,%s)",
               chan.trackNr(),chan.indexNr(),qAbsTime.str());
 
             // first track start?
@@ -705,7 +830,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
 		      chan.trackNr(),qAbsTime.str(),cab);
               if ((qAbsTime.lba()-150)!=info[0].start)
               {
-                message(1, "TOC SAYS TRACK STARTS AT %s!",
+                message(2, "TOC SAYS TRACK STARTS AT %s!",
                   Msf(info[0].start).str());
               }
             }
@@ -716,16 +841,16 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
                 chan.trackNr(),qAbsTime.str());
               if ((qAbsTime.lba()-150)!=info[cat].start)
               {
-                message(1, "TOC SAYS TRACK STARTS AT %s!",
+                message(2, "TOC SAYS TRACK STARTS AT %s!",
 			Msf(info[cat].start).str());
               }
             }
             // pregap of next track?
             else if ((chan.trackNr()==(cat+1)) && (chan.indexNr()==0))
             {
-              message(1, "Found pregap of track %02d with size %s at ",
+              message(2, "Found pregap of track %02d with size %s at ",
 		      chan.trackNr(), qRelTime.str());
-	      message(1, "%s", qAbsTime.str());
+	      message(2, "%s", qAbsTime.str());
 	      if (cat <= lat) {
 		info[cat].pregap = qRelTime.lba();
 	      }
@@ -735,7 +860,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
             {
               if (chan.indexNr() > 1)
               {
-		message(1, "Found index number %02d at %s.",
+		message(2, "Found index number %02d at %s.",
 			chan.indexNr(),qAbsTime.str());
 
 		int indexCnt = info[cat - 1].indexCnt;
@@ -747,7 +872,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
             }
             else if (chan.trackNr()==0xAA)
             {
-              message(1, "LEAD OUT TRACK ENCOUNTERED!");
+              message(2, "LEAD OUT TRACK ENCOUNTERED!");
               n=0;
               i=0;
               rab=0;
@@ -771,7 +896,7 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
 	  // subchannel contains ISRC code
 	  if (info[cat - 1].isrcCode[0] == 0) {
 	    memcpy(info[cat - 1].isrcCode, chan.isrc(), 13);
-	    message(1, "Found ISRC code of track %02d.", cat);
+	    message(2, "Found ISRC code of track %02d.", cat);
 	  }
 	}
       }
@@ -951,7 +1076,8 @@ int PlextorReader::readAudioRangePlextor(int fd,  long startLba, long endLba,
   return 0;
 }
 
-int PlextorReader::readSubChannels(long lba, long len, SubChannel ***chans,
+int PlextorReader::readSubChannels(TrackData::SubChannelMode,
+				   long lba, long len, SubChannel ***chans,
 				   Sample *audioData)
 {
   unsigned char cmd[12];
@@ -976,6 +1102,17 @@ int PlextorReader::readSubChannels(long lba, long len, SubChannel ***chans,
 
     cmdLen = 10;
   }
+  else if (options_ & OPT_PLEX_DAE_D4_12) {
+    cmd[0] = 0xd4;
+    cmd[2] = lba >> 24;
+    cmd[3] = lba >> 16;
+    cmd[4] = lba >> 8;
+    cmd[5] = lba;
+    cmd[8] = len >> 8;
+    cmd[9] = len;
+
+    cmdLen = 12;
+  }
   else {
     cmd[0]=0xd8; // READ CDDA
     cmd[2] = lba >> 24;
@@ -1007,14 +1144,15 @@ int PlextorReader::readSubChannels(long lba, long len, SubChannel ***chans,
   return 0;
 }
 
-int PlextorReader::readAudioRange(int fd, long start, long end,
-				  int startTrack, int endTrack, 
+int PlextorReader::readAudioRange(ReadDiskInfo *rinfo, int fd, long start,
+				  long end, int startTrack, int endTrack, 
 				  TrackInfo *info)
 {
   if (model_ != 0 && (options_ & OPT_PLEX_USE_PARANOIA) == 0) {
     // we have a plextor drive -> use the special plextor method which
     // will also detect pre-gaps, index marks and ISRC codes
-    return readAudioRangePlextor(fd, start, end, startTrack, endTrack, info);
+    return readAudioRangePlextor(rinfo, fd, start, end, startTrack, endTrack,
+				 info);
   }
   
   if (!onTheFly_) {
@@ -1028,12 +1166,19 @@ int PlextorReader::readAudioRange(int fd, long start, long end,
     message(1, "Analyzing...");
     
     for (t = startTrack; t <= endTrack; t++) {
+      long totalProgress;
+
       message(1, "Track %d...", t + 1);
 
+      totalProgress = t * 1000;
+      totalProgress /= rinfo->tracks;
+      sendReadCdProgressMsg(RCD_ANALYZING, rinfo->tracks, t + 1, 0,
+			    totalProgress);
+
 
       if (!fastTocReading_) {
 	if (pregap != 0)
-	  message(1, "Found pre-gap: %s", Msf(pregap).str());
+	  message(2, "Found pre-gap: %s", Msf(pregap).str());
 
 	slba = info[t].start;
 	if (info[t].mode == info[t + 1].mode)
@@ -1063,12 +1208,18 @@ int PlextorReader::readAudioRange(int fd, long start, long end,
       info[t].isrcCode[0] = 0;
       readIsrc(t + 1, info[t].isrcCode);
       if (info[t].isrcCode[0] != 0)
-	message(1, "Found ISRC code.");
+	message(2, "Found ISRC code.");
+
+      totalProgress = (t + 1) * 1000;
+      totalProgress /= rinfo->tracks;
+
+      sendReadCdProgressMsg(RCD_ANALYZING, rinfo->tracks, t + 1, 1000,
+			    totalProgress);
     }
 
     message(1, "Reading...");
   }
 
-  return CdrDriver::readAudioRangeParanoia(fd, start, end, startTrack,
+  return CdrDriver::readAudioRangeParanoia(rinfo, fd, start, end, startTrack,
 					   endTrack, info);
 }
