@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998, 1999  Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2001  Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,43 +16,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
-/*
- * $Log: dao.cc,v $
- * Revision 1.12  1999/05/11 20:04:09  mueller
- * SYSV message queues are not used anymore. The communication between
- * reader and writer is based on a polling mechanism, now.
- * Added posix thread support.
- *
- * Revision 1.11  1999/03/27 20:57:27  mueller
- * Improved buffer code. Does not use semaphores anymore.
- * The reader is now the forked process.
- *
- * Revision 1.10  1999/01/24 16:03:57  mueller
- * Applied Radek Doulik's ring buffer patch. Added some cleanups and
- * improved behavior in error case.
- *
- * Revision 1.9  1998/10/03 14:33:46  mueller
- * Applied patch from Bjoern Fischer <bfischer@Techfak.Uni-Bielefeld.DE>:
- * Cosmetic changes and correction of real time scheduling selection.
- *
- * Revision 1.8  1998/09/22 19:14:29  mueller
- * Added locking of memory pages.
- *
- * Revision 1.7  1998/09/21 19:35:51  mueller
- * Added real time scheduling. Submitted by Bjoern Fischer <bfischer@Techfak.Uni-Bielefeld.DE>.
- *
- * Revision 1.6  1998/09/06 13:34:22  mueller
- * Use 'message()' for printing messages.
- *
- * Revision 1.5  1998/08/30 19:21:15  mueller
- * Rearranged the code.
- *
- * Revision 1.4  1998/08/15 20:47:16  mueller
- * Added support for GenericMMC driver.
- *
- */
-
-static char rcsid[] = "$Id: dao.cc,v 1.12 1999/05/11 20:04:09 mueller Exp mueller $";
 
 #include <config.h>
 
@@ -70,7 +33,6 @@ static char rcsid[] = "$Id: dao.cc,v 1.12 1999/05/11 20:04:09 mueller Exp muelle
 #include <linux/unistd.h>
 #include <linux/types.h>
 #include <linux/sysctl.h>
-#include <sys/sysctl.h>
 #endif
 
 #ifdef HAVE_SYS_MMAN_H
@@ -79,16 +41,21 @@ static char rcsid[] = "$Id: dao.cc,v 1.12 1999/05/11 20:04:09 mueller Exp muelle
 
 #ifdef USE_POSIX_THREADS
 #include <pthread.h>
-#include <sched.h>
 #else
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #endif
 
+#include "dao.h"
+
 #include "port.h"
+#include "util.h"
+
 
 #define DEBUG_WRITE 0
 
+
+
 #if defined(__FreeBSD__)
 #define IPC_ARG_T void
 #else
@@ -96,39 +63,6 @@ static char rcsid[] = "$Id: dao.cc,v 1.12 1999/05/11 20:04:09 mueller Exp muelle
 #endif
 
 
-/* Select POSIX scheduler interface for real time scheduling if possible */
-#if (defined HAVE_SCHED_GETPARAM) && (defined HAVE_SCHED_GET_PRIORITY_MAX) && (defined HAVE_SCHED_SETSCHEDULER) && (!defined LINUX_QNX_SCHEDULING)
-#define POSIX_SCHEDULING
-#endif
-
-#ifdef LINUX_QNX_SCHEDULING
-
-#define SCHED_OTHER     0
-#define SCHED_FIFO      1
-#define SCHED_RR        2
-
-struct sched_param {
-  unsigned int priority;
-  int fork_penalty_threshold;
-  unsigned int starvation_threshold;
-  unsigned int ts_max;
-  unsigned int run_q, run_q_min, run_q_max;
-};
-extern "C" int sched_setparam __P((pid_t __pid, const struct sched_param *__param));
-extern "C" int sched_getparam __P((pid_t __pid, struct sched_param *__param));
-extern "C" int sched_setscheduler __P((pid_t __pid, int __policy, const struct sched_param *__param));
-extern "C" int sched_getscheduler __P((pid_t __pid));
-
-#elif defined POSIX_SCHEDULING
-
-#include <sched.h>
-
-#endif
-
-#include "dao.h"
-#include "util.h"
-#include "remote.h"
-
 struct ShmSegment {
   int id;
   char *buffer;
@@ -140,8 +74,9 @@ struct Buffer {
   TrackData::Mode trackMode; // mode of track may differ from 'mode' if data
                              // blocks must be encoded in audio blocks,
                              // only used for message printing
+  TrackData::SubChannelMode subChanMode; // sub-channel data mode
   int trackNr; // if != 0 a new track with given number has started
-
+  int trackProgress; // reading progress of current track 0..1000
   char *buffer; // address of buffer that should be written
 };
   
@@ -264,9 +199,11 @@ static void unblockSignals()
 // return: 0: OK 
 //         1: child process terminated and has been collected with 'wait()'
 //         2: error -> child process must be terminated
-static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
-		  long remoteMode)
+static int writer(const Toc *toc, CdrDriver *cdr, BufferHeader *header,
+		  long lba, int speed)
 {
+  long total = toc->length().lba() * AUDIO_BLOCK_LEN;
+  long totalTracks = toc->nofTracks();
   long cnt = 0;
   long blkCount = 0;
   long len = 0;
@@ -274,22 +211,17 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
   long lastMb = 0;
   long buffered;
   int buffFill;
+  int writerFill = 0;
   int minFill = 100;
   int maxFill = 0;
   int actTrackNr = 0;
   long actProgress;
   TrackData::Mode dataMode;
-  DaoWritingProgress remoteMsg;
-  char remoteMsgSync[4];
+  TrackData::SubChannelMode subChanMode;
 #ifndef USE_POSIX_THREADS
   int status;
 #endif
 
-  remoteMsgSync[0] = 0xff;
-  remoteMsgSync[1] = 0x00;
-  remoteMsgSync[2] = 0xff;
-  remoteMsgSync[3] = 0x00;
-
   message (3, "Waiting for reader process");
 
   while (header->buffersFilled == 0) {
@@ -320,19 +252,8 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
   message (3, "Awaken, will start writing");
 
   if (cdr != NULL) {
-    if (remoteMode) {
-      remoteMsg.status = 1; // writing lead-in
-      remoteMsg.track = 0;
-      remoteMsg.totalProgress = 0;
-      remoteMsg.bufferFillRate = 100;
-
-      if (write(3/*fileno(stdout)*/, remoteMsgSync, sizeof(remoteMsgSync)) < 0 ||
-	  write(3/*fileno(stdout)*/, (const char*)&remoteMsg,
-		sizeof(remoteMsg)) < 0) {
-	message(-1, "Disabling remote messages.");
-	remoteMode = 0;
-      }
-    }
+    cdr->sendWriteCdProgressMsg(CdrDriver::WCD_LEADIN,
+				totalTracks, 0, 0, 0, 100);
 
     blockSignals();
     if (cdr->startDao() != 0) {
@@ -343,7 +264,7 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
   }
 
   do {
-    //message(3, "Slave: waiting for master.");
+    //message(4, "Slave: waiting for master.");
 
     while (header->buffersWritten == header->buffersRead) {
       if (header->readerTerminated) {
@@ -363,12 +284,13 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
       }
 #endif
 
-      mSleep(20);
+      mSleep(10);
     }
 
     Buffer &buf = header->buffers[header->buffersWritten % header->nofBuffers];
     len = buf.bufLen;
     dataMode = buf.mode;
+    subChanMode = buf.subChanMode;
 
     if (header->readerFinished) {
       buffFill = 100;
@@ -408,21 +330,10 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
 	fclose(fp);
 #endif
 
-      if (remoteMode) {
-	remoteMsg.status = 3; // writing lead-out
-	remoteMsg.track = 0xaa;
-	remoteMsg.totalProgress = 1000;
-	remoteMsg.bufferFillRate = 100;
-
-	if (write(3/*fileno(stdout)*/, remoteMsgSync, sizeof(remoteMsgSync)) < 0 ||
-	    write(3/*fileno(stdout)*/, (const char*)&remoteMsg,
-		  sizeof(remoteMsg)) < 0) {
-	  message(-1, "Disabling remote messages.");
-	  remoteMode = 0;
-	}
-      }
-
       if (cdr != NULL) {
+	cdr->sendWriteCdProgressMsg(CdrDriver::WCD_LEADOUT,
+				    totalTracks, 0xaa, 1000, 1000, 100);
+	
 	blockSignals();
 	if (cdr->finishDao() != 0) {
 	  unblockSignals();
@@ -438,59 +349,70 @@ static int writer(CdrDriver *cdr, long total, BufferHeader *header, long lba,
     blkCount += len;
 
     if (buf.trackNr > 0) {
-      message(1, "Writing track %02d (mode %s/%s)...", buf.trackNr,
+      message(1, "Writing track %02d (mode %s/%s %s)...", buf.trackNr,
 	      TrackData::mode2String(buf.trackMode),
-	      TrackData::mode2String(dataMode));
+	      TrackData::mode2String(dataMode),
+	      TrackData::subChannelMode2String(subChanMode));
 
       actTrackNr = buf.trackNr;
     }
 
-    //message(3, "Slave: writing buffer %p (%ld).", buf, len);
+    //message(4, "Slave: writing buffer %p (%ld).", buf, len);
 
 #if DEBUG_WRITE
     if (fp != NULL) {
-      if (cdr != NULL)
-	fwrite(buf.buffer, cdr->blockSize(dataMode), len, fp);
-      else
+      if (cdr != NULL) {
+	message(0, "dao: blockSize: %ld", cdr->blockSize(dataMode, subChanMode));
+	
+	fwrite(buf.buffer, cdr->blockSize(dataMode, subChanMode), len, fp);
+      }
+      else {
 	fwrite(buf.buffer, 2352, len, fp);
+      }
     }	
 #endif
 
+    // Write track data.
     if (cdr != NULL) {
       blockSignals();
-      if (cdr->writeData(dataMode, lba, buf.buffer, len) != 0) {
+      if (cdr->writeData(dataMode, subChanMode, lba, buf.buffer, len) != 0) {
 	message(-2, "Writing failed - buffer under run?");
 	unblockSignals();
 	return 2;
       }
-      else {
-	cntMb = cnt >> 20;
-	if (cntMb > lastMb) {
-	  message(1, "Wrote %ld of %ld MB (Buffer %3d%%).\r", cnt >> 20,
-		  total >> 20, buffFill);
-	  lastMb = cntMb;
-	}
-      }
-      unblockSignals();
-    }
-    else {
-      message(1, "Read %ld of %ld MB.\r", cnt >> 20, total >> 20);
-    }
 
-    if (remoteMode) {
+      // Print stat line update every megabyte.
+      cntMb = cnt >> 20;
+      if (cntMb > lastMb) {
+        long totalcap, availcap;
+        if (cdr->readBufferCapacity(&totalcap, &availcap)) {
+          writerFill = (int)((1.0 - ((double)availcap / (double)totalcap))
+                             * 100.0);
+          message(1, "Wrote %ld of %ld MB (Buffers %3d%% %3d%%).\n",
+                  cnt >> 20, total >> 20, buffFill, writerFill);
+        } else {
+          message(1, "Wrote %ld of %ld MB (Buffer %3d%%).\n",
+                  cnt >> 20, total >> 20, buffFill);
+        }
+        lastMb = cntMb;
+      }
+
+      unblockSignals();
+
       actProgress = cnt;
       actProgress /= total / 1000;
 
-      remoteMsg.status = 2; // writing data
-      remoteMsg.track = actTrackNr;
-      remoteMsg.totalProgress = actProgress;
-      remoteMsg.bufferFillRate = buffFill;
-      
-      if (write(3/*fileno(stdout)*/, remoteMsgSync, sizeof(remoteMsgSync)) < 0 ||
-	  write(3/*fileno(stdout)*/, (const char*)&remoteMsg,
-		sizeof(remoteMsg)) < 0) {
-	message(-1, "Disabling remote messages.");
-	remoteMode = 0;
+      cdr->sendWriteCdProgressMsg(CdrDriver::WCD_DATA, totalTracks, actTrackNr,
+				  buf.trackProgress, actProgress, buffFill,
+                                  writerFill);
+    }
+    else {
+      if (speed > 0) {
+	message(1, "Read %ld of %ld MB (Buffer %3d%%).\r", cnt >> 20, total >> 20, buffFill);
+	mSleep(1000 / speed);
+      }
+      else {
+	message(1, "Read %ld of %ld MB.\r", cnt >> 20, total >> 20);
       }
     }
 
@@ -526,10 +448,15 @@ static void *reader(void *args)
   int first = header->nofBuffers;
   const Track *track;
   int trackNr = 1;
-  Msf tstart, tend;
   TrackData::Mode dataMode;
-  int encodingMode = 1;
+  TrackData::SubChannelMode subChanMode;
+  int encodingMode = 0;
+  int subChanEncodingMode = 1;
   int newTrack;
+  long tact; // number of blocks already read from current track
+  long tprogress;
+
+  setRealTimeScheduling(4);
 
   if (cdr != NULL) {
     if (cdr->bigEndianSamples() == 0) {
@@ -538,21 +465,27 @@ static void *reader(void *args)
     }
     encodingMode = cdr->encodingMode();
   }
-  message(3, "Swap: %d", swap);
+  message(4, "Swap: %d", swap);
 
   TrackIterator itr(toc);
   TrackReader reader;
 
-  track = itr.first(tstart, tend);
+  track = itr.first();
   reader.init(track);
 
   if (reader.openData() != 0) {
     message(-2, "Opening of track data failed.");
     goto fail;
   }
+
   newTrack = 1;
+  tact = 0;
 
   dataMode = (encodingMode == 0) ? TrackData::AUDIO : track->type();
+  subChanMode = track->subChannelType();
+
+  if (cdr != NULL)
+    subChanEncodingMode = cdr->subChannelEncodingMode(subChanMode);
 
   do {
     n = (length > BUFFER_SIZE ? BUFFER_SIZE : length);
@@ -560,7 +493,8 @@ static void *reader(void *args)
     Buffer &buf = header->buffers[header->buffersRead % header->nofBuffers];
 
     do {
-      rn = reader.readData(encodingMode, lba, buf.buffer, n);
+      rn = reader.readData(encodingMode, subChanEncodingMode, lba, buf.buffer,
+			   n);
     
       if (rn < 0) {
 	message(-2, "Reading of track data failed.");
@@ -568,40 +502,53 @@ static void *reader(void *args)
       }
       
       if (rn == 0) {
-	track = itr.next(tstart, tend);
+	track = itr.next();
 	reader.init(track);
 
 	if (reader.openData() != 0) {
 	  message(-2, "Opening of track data failed.");
 	  goto fail;
 	}
+
 	trackNr++;
+
 	if (encodingMode != 0)
 	  dataMode = track->type();
 
+	subChanMode = track->subChannelType();
+
+	if (cdr != NULL)
+	  subChanEncodingMode = cdr->subChannelEncodingMode(subChanMode);
+
 	newTrack = 1;
+	tact = 0;
       }
     } while (rn == 0);
 
     lba += rn;
+    tact += rn;
 
-    if (track->type() == TrackData::AUDIO) {
-      if (swap) {
-	// swap audio data
-	swapSamples((Sample *)(buf.buffer), rn * SAMPLES_PER_BLOCK);
-      }
-    }
-    else {
-      if (encodingMode == 0 && cdr != NULL && cdr->bigEndianSamples() == 0) {
-	// swap encoded data blocks
-	swapSamples((Sample *)(buf.buffer), rn * SAMPLES_PER_BLOCK);
-      }
+    if (cdr != NULL &&
+	((track->type() == TrackData::AUDIO && swap) ||
+	 (encodingMode == 0 && cdr->bigEndianSamples() == 0))) {
+      // swap audio data 
+      long blockLen = cdr->blockSize(dataMode, subChanMode);
+      char *brun = buf.buffer;
+      int i;
+	
+      for (i = 0; i < rn; i++, brun += blockLen)
+	swapSamples((Sample *)brun, SAMPLES_PER_BLOCK);
     }
 
-    // notify writer that it can write buffer
     buf.bufLen = rn;
     buf.mode = dataMode;
     buf.trackMode = track->type();
+    buf.subChanMode = subChanMode;
+
+    tprogress = tact * 1000;
+    tprogress /= track->length().lba();
+
+    buf.trackProgress = tprogress;
 
     if (newTrack) {
       // inform write process that it should print message about new track
@@ -625,11 +572,11 @@ static void *reader(void *args)
     }
     
     // wait for writing process to finish writing of previous buffer
-    //message(3, "Reader: waiting for Writer.");
+    //message(4, "Reader: waiting for Writer.");
     while (header->buffersRead - header->buffersWritten 
 	   == header->nofBuffers &&
 	   header->terminateReader == 0) {
-      mSleep(20);
+      mSleep(10);
     }
 
 
@@ -667,21 +614,14 @@ fail:
 
 
 int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
-		    int testMode, int remoteMode)
+		    int testMode, int speed)
 {
-  long length = toc->length().lba();
-  long total = length * AUDIO_BLOCK_LEN;
   int err = 0;
   BufferHeader *header = NULL;
   long nofShmSegments = 0;
   ShmSegment *shmSegments = NULL;
   long startLba = 0;
 
-#if defined(POSIX_SCHEDULING) || defined(LINUX_QNX_SCHEDULING) || \
-    defined(USE_POSIX_THREADS)
-  struct sched_param schedp;
-#endif
-
 #ifdef USE_POSIX_THREADS
   pthread_t readerThread;
   pthread_attr_t readerThreadAttr;
@@ -698,23 +638,6 @@ int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
   }
 #endif
 
-  if (remoteMode) {
-    // switch stdout to non blocking IO
-    int flags;
-
-    if ((flags = fcntl(3/*fileno(stdout)*/, F_GETFL)) == -1) {
-      message(-2, "Cannot get flags of stdout: %s", strerror(errno));
-      return 1;
-    }
-
-    flags |= O_NONBLOCK;
-
-    if (fcntl(3/*fileno(stdout)*/, F_SETFL, flags) < 0) {
-      message(-2, "Cannot set flags of stdout: %s", strerror(errno));
-      return 1;
-    }
-  }
-
   if (getSharedMemory(nofBuffers, &header, &nofShmSegments,
 		      &shmSegments)  != 0) {
     releaseSharedMemory(nofShmSegments, shmSegments);
@@ -757,29 +680,6 @@ int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
     err = 1; goto fail;
   }
   
-
-#if defined(POSIX_SCHEDULING) && defined(HAVE_PTHREAD_ATTR_SETSCHEDPOLICY) && \
-  defined(HAVE_PTHREAD_ATTR_SETSCHEDPARAM)
-
-  if (geteuid() == 0) {
-    if (pthread_attr_setschedpolicy(&readerThreadAttr, SCHED_RR) == 0) {
-      struct sched_param schedParam;
-
-      if (pthread_attr_getschedparam(&readerThreadAttr, &schedParam) == 0) {
-	schedParam.sched_priority = sched_get_priority_max(SCHED_RR) - 4;
-	if (pthread_attr_setschedparam(&readerThreadAttr, &schedParam) != 0) {
-	  message(-1, "pthread_attr_setschedparam failed: %s",
-		  strerror(errno));
-	}
-      }
-    }
-  }
-#else
-
-  message(-1, "Real time scheduling is not available for reader thread.");
-
-#endif
-
   ReaderArgs rargs;
 
   rargs.toc = toc;
@@ -795,8 +695,6 @@ int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
   }
   else {
     threadStarted = 1;
-
-    
   }
 
 #else /* USE_POSIX_THREADS */
@@ -806,34 +704,12 @@ int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
 
     setsid(); // detach from controlling terminal
 
-#ifdef LINUX_QNX_SCHEDULING
-
-    if (geteuid() == 0) {
-      sched_getparam (0, &schedp);
-      schedp.run_q_min = schedp.run_q_max = 1;
-      if (sched_setscheduler (0, SCHED_RR, &schedp) != 0) {
-	message(-1, "Cannot setup real time scheduling: %s", strerror(errno));
-      }
-    }
-
-#elif defined POSIX_SCHEDULING
-
-    if (geteuid() == 0) {
-      sched_getparam (0, &schedp);
-      schedp.sched_priority = sched_get_priority_max (SCHED_RR) - 4;
-      if (sched_setscheduler (0, SCHED_RR, &schedp) != 0) {
-	message(-1, "Cannot setup real time scheduling: %s", strerror(errno));
-      }
-    }
-
-#endif
-
 #ifdef HAVE_MLOCKALL
     if (geteuid() == 0) {
       if (mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
 	message(-1, "Cannot lock memory pages: %s", strerror(errno));
       }
-      message(3, "Reader process memory locked");
+      message(4, "Reader process memory locked");
     }
 #endif
 
@@ -853,74 +729,25 @@ int writeDiskAtOnce(const Toc *toc, CdrDriver *cdr, int nofBuffers, int swap,
   }
 #endif /* USE_POSIX_THREADS */
 
-  // setup scheduler for writing process/thread
-#if defined(USE_POSIX_THREADS)
-
-#if defined(HAVE_PTHREAD_GETSCHEDPARAM) && defined(HAVE_PTHREAD_SETSCHEDPARAM)
-  if (geteuid() == 0) {
-    int schedPolicy;
-
-    pthread_getschedparam(pthread_self(), &schedPolicy, &schedp);
-    schedp.sched_priority = sched_get_priority_max(SCHED_RR) - 5;
-
-    if (pthread_setschedparam(pthread_self(), SCHED_RR, &schedp) != 0) {
-      message(-1, "Cannot setup real time scheduling: %s", strerror(errno));
-    }
-    else {
-      message(1, "Using POSIX real time scheduling.");
-    }
+  switch (setRealTimeScheduling(5)) {
+  case 1:
+    message(-1, "No super user permission to setup real time scheduling.");
+    break;
+  case 2:
+    message(2, "Real time scheduling not available.");
+    break;
   }
-  else {
-    message(-1, "No root permissions for real time scheduling.");
-  }
-#else
-  message(-1, "Real time scheduling is not available.");
-#endif
-
-#elif defined(LINUX_QNX_SCHEDULING)
-  
-  if (geteuid() == 0) {
-    sched_getparam (0, &schedp);
-    schedp.run_q_min = schedp.run_q_max = 2;
-    if (sched_setscheduler (0, SCHED_RR, &schedp) != 0) {
-      message(-1, "Cannot setup real time scheduling: %s", strerror(errno));
-    }
-    else {
-      message(1, "Using Linux QNX real time scheduling.");
-    }
-  }
-  else {
-    message(-1, "No root permissions for real time scheduling.");
-  }
-
-#elif defined POSIX_SCHEDULING
-  
-  if (geteuid() == 0) {
-    sched_getparam (0, &schedp);
-    schedp.sched_priority = sched_get_priority_max (SCHED_RR) - 5;
-    if (sched_setscheduler (0, SCHED_RR, &schedp) != 0) {
-      message(-1, "Cannot setup real time scheduling: %s", strerror(errno));
-    }
-    else {
-      message(1, "Using POSIX real time scheduling.");
-    }
-  }
-  else {
-    message(-1, "No root permissions for real time scheduling.");
-  }
-  
-#endif
 
 #ifdef HAVE_MLOCKALL
   if (geteuid() == 0) {
     if (mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
       message(-1, "Cannot lock memory pages: %s", strerror(errno));
     }
-    message(3, "Memory locked");
+    message(4, "Memory locked");
   }
 #endif
 
-  switch (writer(cdr, total, header, startLba, remoteMode)) {
+  switch (writer(toc, cdr, header, startLba, speed)) {
   case 1: // error, reader process terminated abnormally
 #ifndef USE_POSIX_THREADS
     pid = 0;
@@ -979,7 +806,7 @@ static int getSharedMemory(long nofBuffers,
 			   ShmSegment **shmSegment)
 {
   long b;
-  long bufferSize = BUFFER_SIZE * AUDIO_BLOCK_LEN;
+  long bufferSize = BUFFER_SIZE * (AUDIO_BLOCK_LEN + PW_SUBCHANNEL_LEN);
 
   *header = NULL;
   *nofSegments = 0;
@@ -993,16 +820,16 @@ static int getSharedMemory(long nofBuffers,
   *nofSegments = 1;
 
   (*shmSegment)->id = -1;
+
   (*shmSegment)->buffer = new char[sizeof(BufferHeader) +
-				   nofBuffers * sizeof(Buffer) +
-				   nofBuffers * bufferSize];
+				  nofBuffers * sizeof(Buffer) +
+				  nofBuffers * bufferSize];
 
   if ( (*shmSegment)->buffer == NULL) {
     message(-2, "Cannot allocated memory for ring buffer.");
     return 1;
   }
 
-
   *header = (BufferHeader*)((*shmSegment)->buffer);
   (*header)->nofBuffers = nofBuffers;
   (*header)->buffers =
@@ -1037,7 +864,7 @@ static int getSharedMemory(long nofBuffers,
 			   ShmSegment **shmSegments)
 {
   long i, b;
-  long bufferSize = BUFFER_SIZE * AUDIO_BLOCK_LEN;
+  long bufferSize = BUFFER_SIZE * (AUDIO_BLOCK_LEN + PW_SUBCHANNEL_LEN);
   long maxSegmentSize = 0;
   long bcnt = 0;
 
@@ -1068,7 +895,7 @@ static int getSharedMemory(long nofBuffers,
   maxSegmentSize = 1 * 1024 * 1024; // 1 MB
 #endif
 
-  message(3, "Shm max segement size: %ld (%ld MB)", maxSegmentSize,
+  message(4, "Shm max segement size: %ld (%ld MB)", maxSegmentSize,
 	  maxSegmentSize >> 20);
 
   if (maxSegmentSize < sizeof(BufferHeader) + nofBuffers * sizeof(Buffer)) {
@@ -1092,7 +919,7 @@ static int getSharedMemory(long nofBuffers,
 
   *shmSegments = new ShmSegment[*nofSegments];
 
-  message(3, "Using %ld shared memory segments.", *nofSegments);
+  message(4, "Using %ld shared memory segments.", *nofSegments);
 
   for (i = 0; i < *nofSegments; i++) {
     (*shmSegments)[i].id = -1;
@@ -1113,7 +940,7 @@ static int getSharedMemory(long nofBuffers,
       segmentLength += sizeof(BufferHeader) + nofBuffers * sizeof(Buffer);
     }
 
-    (*shmSegments)[i].id = shmget(IPC_PRIVATE, segmentLength, 0600);
+    (*shmSegments)[i].id = shmget(IPC_PRIVATE, segmentLength, 0600|IPC_CREAT);
     if ((*shmSegments)[i].id < 0) {
       message(-2, "Cannot create shared memory segment: %s",
 	      strerror(errno));
