@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998, 1999 Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2002 Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,14 +16,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
-/*
- * $Log: CdDevice.cc,v $
- * Revision 1.1  1999/09/03 16:05:14  mueller
- * Initial revision
- *
- */
-
-static char rcsid[] = "$Id: CdDevice.cc,v 1.1 1999/09/03 16:05:14 mueller Exp mueller $";
 
 #include <sys/time.h>
 #include <sys/types.h>
@@ -32,26 +24,29 @@ static char rcsid[] = "$Id: CdDevice.cc,v 1.1 1999/09/03 16:05:14 mueller Exp mu
 #include <unistd.h>
 #include <malloc.h>
 #include <errno.h>
-#include <string.h>
 #include <ctype.h>
 #include <assert.h>
 
+#include <gtkmm.h>
+#include <gnome.h>
+
 #include "TocEdit.h"
 #include "CdDevice.h"
 #include "ProcessMonitor.h"
 #include "xcdrdao.h"
 #include "guiUpdate.h"
-#include "RecordProgressDialog.h"
+#include "ProgressDialog.h"
 #include "Settings.h"
 
+#include "config.h"
 #include "remote.h"
 #include "ScsiIf.h"
 #include "CdrDriver.h"
-
 #include "util.h"
 #include "Toc.h"
 
-#define DRIVER_IDS 12
+#define DRIVER_IDS 13
+#define DRIVER_ID_DEFAULT 2
 
 CdDevice *CdDevice::DEVICE_LIST_ = NULL;
 
@@ -67,27 +62,23 @@ char *CdDevice::DRIVER_NAMES_[DRIVER_IDS] = {
   "sony-cdu948",
   "taiyo-yuden",
   "teac-cdr55",
+  "toshiba",
   "yamaha-cdr10x"
 };
   
 
-CdDevice::CdDevice(int bus, int id, int lun, const char *vendor,
-		   const char *product)
+CdDevice::CdDevice(const char* dev, const char *vendor, const char *product)
 {
-  bus_ = bus;
-  id_ = id;
-  lun_ = lun;
-
-  vendor_ = strdupCC(vendor);
-  product_ = strdupCC(product);
+  dev_ = dev;
+  vendor_ = vendor;
+  product_ = product;
   
   driverId_ = 0; // undefined
   options_ = 0;
-  specialDevice_ = NULL;
 
   deviceType_ = CD_R;
 
-  manuallyConfigured_ = 0;
+  manuallyConfigured_ = false;
 
   status_ = DEV_UNKNOWN;
 
@@ -95,9 +86,12 @@ CdDevice::CdDevice(int bus, int id, int lun, const char *vendor,
 
   progressStatusChanged_ = 0;
   progressStatus_ = 0;
+  progressTotalTracks_ = 0;
   progressTrack_ = 0;
   progressTotal_ = 0;
+  progressTrackRelative_ = 0;
   progressBufferFill_ = 0;
+  progressWriterFill_ = 0;
 
   process_ = NULL;
 
@@ -105,21 +99,13 @@ CdDevice::CdDevice(int bus, int id, int lun, const char *vendor,
   scsiIfInitFailed_ = 0;
 
   next_ = NULL;
+  slaveDevice_ = NULL;
 
   autoSelectDriver();
 }
 
 CdDevice::~CdDevice()
 {
-  delete[] vendor_;
-  vendor_ = NULL;
-
-  delete[] product_;
-  product_ = NULL;
-
-  delete[] specialDevice_;
-  specialDevice_ = NULL;
-
   delete scsiIf_;
   scsiIf_ = NULL;
 }
@@ -127,11 +113,9 @@ CdDevice::~CdDevice()
 char *CdDevice::settingString() const
 {
   char buf[100];
-  string s;
+  std::string s;
 
-  sprintf(buf, "%d,%d,%d,'", bus_, id_, lun_);
-
-  s += buf;
+  s = "'" + dev_ + "','";
   s += vendor_;
   s += "','";
   s += product_;
@@ -155,91 +139,18 @@ char *CdDevice::settingString() const
 
   s += ",";
 
-  sprintf(buf, "0x%lx,", options_);
+  sprintf(buf, "0x%lx", options_);
   s += buf;
 
-  s += "'";
-  if (specialDevice_ != NULL)
-    s += specialDevice_;
-  s += "'";
-
   return strdupCC(s.c_str());
 }
 
-int CdDevice::bus() const
-{
-  return bus_;
-}
-
-int CdDevice::id() const
-{
-  return id_;
-}
-
-int CdDevice::lun() const
-{
-  return lun_;
-}
-
-const char *CdDevice::vendor() const
-{
-  return vendor_;
-}
-
-const char *CdDevice::product() const
-{
-  return product_;
-}
-
-int CdDevice::driverId() const
-{
-  return driverId_;
-}
-
 void CdDevice::driverId(int id)
 {
   if (id >= 0 && id < DRIVER_IDS) 
     driverId_ = id;
 }
 
-const char *CdDevice::specialDevice() const
-{
-  return specialDevice_;
-}
-
-void CdDevice::specialDevice(const char *s)
-{
-  delete[] specialDevice_;
-
-  if (s == NULL || *s == 0)
-    specialDevice_ = NULL;
-  else
-    specialDevice_ = strdupCC(s);
-
-  // remove existing SCSI interface class, it will be recreated with
-  // respect to special device setting automatically later
-  delete scsiIf_;
-  scsiIf_ = NULL;
-  scsiIfInitFailed_ = 0;
-
-  status_ = DEV_UNKNOWN;
-}
-
-int CdDevice::manuallyConfigured() const
-{
-  return manuallyConfigured_;
-}
-
-void CdDevice::manuallyConfigured(int f)
-{
-  manuallyConfigured_ = (f != 0) ? 1 : 0;
-}
-
-CdDevice::Status CdDevice::status() const
-{
-  return status_;
-}
-
 void CdDevice::status(Status s)
 {
   status_ = s;
@@ -255,28 +166,18 @@ int CdDevice::autoSelectDriver()
   unsigned long options = 0;
   const char *driverName;
 
-  driverName = CdrDriver::selectDriver(1, vendor_, product_, &options);
+  driverName = CdrDriver::selectDriver(1, vendor_.c_str(), product_.c_str(),
+                                       &options);
 
-  if (driverName == NULL) {
-    // select read driver 
-    driverName = CdrDriver::selectDriver(0, vendor_, product_, &options);
-
-    if (driverName != NULL)
-      deviceType_ = CD_ROM;
-  }
-
-  if (driverName != NULL) {
+  if (driverName) {
     driverId_ = driverName2Id(driverName);
     options_ = options;
-
-    return 1;
-  }
-  else {
-    driverId_ = 0;
+  } else {
+    driverId_ = DRIVER_ID_DEFAULT;
     options_ = 0;
-
-    return 0;
   }
+
+  return 1;
 }
 
 int CdDevice::updateStatus()
@@ -292,6 +193,11 @@ int CdDevice::updateStatus()
 
       PROCESS_MONITOR->remove(process_);
       process_ = NULL;
+
+      if (slaveDevice_ != NULL) {
+	slaveDevice_->status(DEV_UNKNOWN);
+	slaveDevice_ = NULL;
+      }
     }
   }
 
@@ -329,20 +235,19 @@ int CdDevice::updateStatus()
   return 0;
 }
 
-int CdDevice::updateProgress()
+bool CdDevice::updateProgress(Glib::IOCondition cond, int fd)
 {
   static char msgSync[4] = { 0xff, 0x00, 0xff, 0x00 };
-  int fd;
   fd_set fds;
   int state = 0;
   char buf[10];
-  DaoWritingProgress msg;
   struct timeval timeout = { 0, 0 };
 
-  //message(0, "Rcvd msg");
+  if (process_ == NULL)
+    return false;
 
-  if (process_ == NULL || (fd = process_->commFd()) < 0)
-    return 0;
+  if (!(cond & Gdk::INPUT_READ))
+    return false;
 
   FD_ZERO(&fds);
   FD_SET(fd, &fds);
@@ -357,7 +262,7 @@ int CdDevice::updateProgress()
     while (state < 4) {
       if (read(fd, buf, 1) != 1) {
 	//message(-2, "Reading of msg sync failed.");
-	return (progressStatusChanged_ != 0) ? 1 : 0;
+	return false;
       }
 
       if (buf[0] == msgSync[state]) {
@@ -372,30 +277,37 @@ int CdDevice::updateProgress()
       }
     }
 
-    if (read(fd, (char *)&msg, sizeof(msg)) == sizeof(msg)) {
-      /*
-	message(0, "progress: %d %d %d %d", msg.status, msg.track,
-	msg.totalProgress, msg.bufferFillRate);
-	*/
-      if (msg.status >= 1 && msg.status <= 3 &&
-	  msg.track >= 0 && msg.track <= 0xaa &&
+    ProgressMsg msg;
+
+    int msgsize = read(fd, (char *)&msg, sizeof(msg));
+    if (msgsize >= PSGMSG_MINSIZE) {
+      if (msg.status >= PGSMSG_MIN && msg.status <= PGSMSG_MAX &&
+	  msg.track >= 0 &&
 	  msg.totalProgress >= 0 && msg.totalProgress <= 1000 &&
 	  msg.bufferFillRate >= 0 && msg.bufferFillRate <= 100) {
 	progressStatus_ = msg.status;
+	progressTotalTracks_ = msg.totalTracks;
 	progressTrack_ = msg.track;
+	progressTrackRelative_ = msg.trackProgress;
 	progressTotal_ = msg.totalProgress;
 	progressBufferFill_ = msg.bufferFillRate;
-
+        if (msgsize == sizeof(msg))
+          progressWriterFill_ = msg.writerFillRate;
+        else
+          progressWriterFill_ = 0;
+	
 	progressStatusChanged_ = 1;
       }
     }
     else {
-      message(-1, "Reading of msg failed.");
+      message(-1, _("Reading of progress message failed."));
     }
-
   }
 
-  return (progressStatusChanged_ != 0) ? 1 : 0;
+  if (progressStatusChanged_)
+    guiUpdate(UPD_PROGRESS_STATUS);
+
+  return true;
 }
 
 CdDevice::DeviceType CdDevice::deviceType() const
@@ -418,35 +330,42 @@ void CdDevice::driverOptions(unsigned long o)
   options_ = o;
 }
 
-// Starts a 'cdrdao' for recording given toc.
-// Return: 0: OK, process succesfully launched
-//         1: error occured
-int CdDevice::recordDao(TocEdit *tocEdit, int simulate, int multiSession,
-			int speed, int eject, int reload)
+// Starts a 'cdrdao' for recording given toc. Returns false if an
+// error occured and the process was not successfully launched.
+bool CdDevice::recordDao(Gtk::Window& parent, TocEdit *tocEdit, int simulate,
+                        int multiSession, int speed, int eject, int reload,
+                        int buffer, int overburn)
 {
-  char *tocFileName;
-  char *args[20];
+  char tocFileName[30];
+  char *args[30];
   int n = 0;
   char devname[30];
   char drivername[50];
   char speedbuf[20];
   char *execName;
   const char *s;
+  char bufferbuf[20];
+  int remoteFdArgNum = 0;
 
-  if (status_ != DEV_READY || process_ != NULL)
-    return 1;
+  if ((status_ != DEV_READY && status_ != DEV_FAULT && status_ != DEV_UNKNOWN)
+      || process_ != NULL)
+    return false;
 
-  if ((tocFileName = tempnam(NULL, "toc")) == NULL) {
-    message(-2, "Cannot create temporary toc-file: %s", strerror(errno));
-    return 1;
+  sprintf(tocFileName, "/tmp/gcdm.toc.XXXXXX");
+  int fd = mkstemp(tocFileName);
+  if (!fd) {
+    message(-2, _("Cannot create temporary toc-file: %s"), strerror(errno));
+    return false;
   }
 
-  if (tocEdit->toc()->write(tocFileName) != 0) {
-    message(-2, "Cannot write temporary toc-file.");
-    return 1;
+  if (!tocEdit->toc()->write(fd)) {
+    close(fd);
+    message(-2, _("Cannot write temporary toc-file."));
+    return false;
   }
 
-  if ((s = SETTINGS->getString(SET_CDRDAO_PATH)) != NULL)
+  close(fd);
+  if ((s = gnome_config_get_string(SET_CDRDAO_PATH)) != NULL)
     execName = strdupCC(s);
   else
     execName = strdupCC("cdrdao");
@@ -461,6 +380,9 @@ int CdDevice::recordDao(TocEdit *tocEdit, int simulate, int multiSession,
 
   args[n++] = "--remote";
 
+  remoteFdArgNum = n;
+  args[n++] = NULL;
+
   args[n++] = "-v0";
 
   if (multiSession)
@@ -478,15 +400,11 @@ int CdDevice::recordDao(TocEdit *tocEdit, int simulate, int multiSession,
   if (reload)
     args[n++] = "--reload";
 
-  args[n++] = "--device";
+  if (overburn)
+    args[n++] = "--overburn";
 
-  if (specialDevice_ != NULL && *specialDevice_ != 0) {
-    args[n++] = specialDevice_;
-  }
-  else {
-    sprintf(devname, "%d,%d,%d", bus_, id_, lun_);
-    args[n++] = devname;
-  }
+  args[n++] = "--device";
+  args[n++] = (char*)dev_.c_str();
 
   if (driverId_ > 0) {
     sprintf(drivername, "%s:0x%lx", driverName(driverId_), options_);
@@ -494,48 +412,45 @@ int CdDevice::recordDao(TocEdit *tocEdit, int simulate, int multiSession,
     args[n++] = drivername;
   }
 
+  if (buffer >= 10) {
+    sprintf(bufferbuf, "%i", buffer);
+    args[n++] = "--buffers";
+    args[n++] = bufferbuf;
+  }
+
   args[n++] = tocFileName;
 
   args[n++] = NULL;
   
   assert(n <= 20);
   
-  int i;
-
-  message(0, "Starting: ");
-  for (i = 0; i < n - 1; i++)
-    message(0, "%s ", args[i]);
-  message(0, "");
-
-  RECORD_PROGRESS_POOL->start(this, tocEdit);
+  PROGRESS_POOL->start(parent, this, tocEdit->filename());
 
   // Remove the SCSI interface of this device to avoid problems with double
   // usage of device nodes.
   delete scsiIf_;
   scsiIf_ = NULL;
 
-  process_ = PROCESS_MONITOR->start(execName, args);
-
-  delete[] execName;
+  process_ = PROCESS_MONITOR->start(execName, args, remoteFdArgNum);
 
+  delete execName;
   if (process_ != NULL) {
     status_ = DEV_RECORDING;
-    free(tocFileName);
+    action_ = A_RECORD;
 
-    /*
     if (process_->commFd() >= 0) {
-      connect_to_method(Gtk_Main::instance()->input(process_->commFd(),
-						    GDK_INPUT_READ),
-			this, &CdDevice::updateProgress);
+        Glib::signal_io().connect(bind(slot(*this, &CdDevice::updateProgress),
+                                       process_->commFd()),
+                                  process_->commFd(),
+                                  Glib::IO_IN | Glib::IO_HUP);
     }
-    */
 
-    return 0;
+    return true;
   }
   else {
     unlink(tocFileName);
     free(tocFileName);
-    return 1;
+    return false;
   }
 }
 
@@ -556,13 +471,386 @@ int CdDevice::progressStatusChanged()
   return 0;
 }
 
-void CdDevice::recordProgress(int *status, int *track, int *totalProgress,
-		      int *bufferFill) const
+void CdDevice::progress(int *status, int *totalTracks, int *track,
+			int *trackProgress, int *totalProgress,
+			int *bufferFill, int *writerFill) const
 {
   *status = progressStatus_;
+  *totalTracks = progressTotalTracks_;
   *track = progressTrack_;
+  *trackProgress = progressTrackRelative_;
   *totalProgress = progressTotal_;
   *bufferFill = progressBufferFill_;
+  *writerFill = progressWriterFill_;
+}
+
+// Starts a 'cdrdao' for reading whole cd.
+// Return: 0: OK, process succesfully launched
+//         1: error occured
+int CdDevice::extractDao(Project& parent, const char *tocFileName,
+                         int correction, int readSubChanMode)
+{
+  char *args[30];
+  int n = 0;
+  char devname[30];
+  char drivername[50];
+  char *execName;
+  const char *s; 
+  char correctionbuf[20];
+  int remoteFdArgNum = 0;
+
+  if ((status_ != DEV_READY && status_ != DEV_FAULT && status_ != DEV_UNKNOWN)
+      || process_ != NULL)
+    return 1;
+
+  if ((s = gnome_config_get_string(SET_CDRDAO_PATH)) != NULL)
+    execName = strdupCC(s);
+  else
+    execName = strdupCC("cdrdao");
+
+
+  args[n++] = execName;
+
+  args[n++] = "read-cd";
+
+  args[n++] = "--remote";
+
+  remoteFdArgNum = n;
+  args[n++] = NULL;
+
+  args[n++] = "-v0";
+
+  args[n++] = "--read-raw";
+
+  switch (readSubChanMode) {
+  case 1:
+    args[n++] = "--read-subchan";
+    args[n++] = "rw";
+    break;
+
+  case 2:
+    args[n++] = "--read-subchan";
+    args[n++] = "rw_raw";
+    break;
+  }
+
+  args[n++] = "--device";
+  args[n++] = (char*)dev_.c_str();
+
+  if (driverId_ > 0) {
+    sprintf(drivername, "%s:0x%lx", driverName(driverId_), options_);
+    args[n++] = "--driver";
+    args[n++] = drivername;
+  }
+
+  sprintf(correctionbuf, "%d", correction);
+  args[n++] = "--paranoia-mode";
+  args[n++] = correctionbuf;
+
+  args[n++] = "--datafile";
+  args[n++] = g_strdup_printf("%s.bin", tocFileName);
+
+  args[n++] = g_strdup_printf("%s.toc", tocFileName);
+
+  args[n++] = NULL;
+  
+  assert(n <= 20);
+  
+  PROGRESS_POOL->start(parent, this, tocFileName, false, false);
+
+  // Remove the SCSI interface of this device to avoid problems with double
+  // usage of device nodes.
+  delete scsiIf_;
+  scsiIf_ = NULL;
+
+  process_ = PROCESS_MONITOR->start(execName, args, remoteFdArgNum);
+
+  delete[] execName;
+
+  if (process_ != NULL) {
+    status_ = DEV_READING;
+    action_ = A_READ;
+
+    if (process_->commFd() >= 0) {
+        Glib::signal_io().connect(bind(slot(*this, &CdDevice::updateProgress),
+                                       process_->commFd()),
+                                  process_->commFd(),
+                                  Glib::IO_IN | Glib::IO_PRI |
+                                  Glib::IO_ERR | Glib::IO_HUP);
+    }
+    return 0;
+  }
+  else {
+    return 1;
+  }
+}
+
+
+void CdDevice::abortDaoReading()
+{
+  if (process_ != NULL && !process_->exited()) {
+    PROCESS_MONITOR->stop(process_);
+  }
+}
+
+// Starts a 'cdrdao' for duplicating a CD.
+// Return: 0: OK, process succesfully launched
+//         1: error occured
+int CdDevice::duplicateDao(Project& parent, int simulate, int multiSession,
+                           int speed, int eject, int reload, int buffer,
+                           int onthefly, int correction, int readSubChanMode, 
+			   CdDevice *readdev)
+{
+  char *args[30];
+  int n = 0;
+  char devname[30];
+  char drivername[50];
+  char r_drivername[50];
+  char speedbuf[20];
+  char correctionbuf[20];
+  char *execName;
+  const char *s;
+  char bufferbuf[20];
+  int remoteFdArgNum = 0;
+
+
+  int rdstat = readdev->status();
+  if ((rdstat != DEV_READY && rdstat != DEV_UNKNOWN && rdstat != DEV_FAULT) ||
+      readdev->process() != NULL)
+    return 1;
+
+  if ((status_ != DEV_READY && status_ != DEV_FAULT && status_ != DEV_UNKNOWN)
+      || process_ != NULL)
+    return 1;
+
+  if ((s = gnome_config_get_string(SET_CDRDAO_PATH)) != NULL)
+    execName = strdupCC(s);
+  else
+    execName = strdupCC("cdrdao");
+
+  args[n++] = execName;
+
+  args[n++] = "copy";
+
+  if (simulate)
+    args[n++] = "--simulate";
+
+  args[n++] = "--remote";
+
+  remoteFdArgNum = n;
+  args[n++] = NULL;
+
+  args[n++] = "-v0";
+
+  if (multiSession)
+    args[n++] = "--multi";
+
+  sprintf(correctionbuf, "%d", correction);
+  args[n++] = "--paranoia-mode";
+  args[n++] = correctionbuf;
+
+  if (speed > 0) {
+    sprintf(speedbuf, "%d", speed);
+    args[n++] = "--speed";
+    args[n++] = speedbuf;
+  }
+
+  if (eject)
+    args[n++] = "--eject";
+
+  if (reload)
+    args[n++] = "--reload";
+
+  if (onthefly)
+    args[n++] = "--on-the-fly";
+
+  switch (readSubChanMode) {
+  case 1:
+    args[n++] = "--read-subchan";
+    args[n++] = "rw";
+    break;
+
+  case 2:
+    args[n++] = "--read-subchan";
+    args[n++] = "rw_raw";
+    break;
+  }
+
+  args[n++] = "--device";
+  args[n++] = (char*)dev_.c_str();
+
+  if (driverId_ > 0) {
+    sprintf(drivername, "%s:0x%lx", driverName(driverId_), options_);
+    args[n++] = "--driver";
+    args[n++] = drivername;
+  }
+
+
+  if (readdev != this) { // reader and write the same, skip source device
+		  
+    args[n++] = "--source-device";
+    args[n++] = (char*)readdev->dev();
+
+    if (readdev->driverId() > 0) {
+      sprintf(r_drivername, "%s:0x%lx", driverName(readdev->driverId()),
+      			 readdev->driverOptions());
+      args[n++] = "--source-driver";
+      args[n++] = r_drivername;
+    }
+  }
+  if (buffer >= 10) {
+    sprintf(bufferbuf, "%i", buffer);
+    args[n++] = "--buffers";
+    args[n++] = bufferbuf;
+  }
+
+
+  args[n++] = NULL;
+  
+  assert(n <= 25);
+  
+  PROGRESS_POOL->start(parent, this, _("CD to CD copy"));
+
+  // Remove the SCSI interface of this device to avoid problems with double
+  // usage of device nodes.
+  delete scsiIf_;
+  scsiIf_ = NULL;
+
+  process_ = PROCESS_MONITOR->start(execName, args, remoteFdArgNum);
+
+  delete[] execName;
+
+  if (process_ != NULL) {
+    slaveDevice_ = readdev;
+    slaveDevice_->status(DEV_READING);
+    status_ = DEV_RECORDING;
+
+    action_ = A_DUPLICATE;
+
+    if (process_->commFd() >= 0) {
+        Glib::signal_io().connect(bind(slot(*this, &CdDevice::updateProgress),
+                                       process_->commFd()),
+                                  process_->commFd(),
+                                  Glib::IO_IN | Glib::IO_HUP);
+    }
+
+    return 0;
+  }
+  else {
+    return 1;
+  }
+}
+
+void CdDevice::abortDaoDuplication()
+{
+  if (process_ != NULL && !process_->exited()) {
+    PROCESS_MONITOR->stop(process_);
+  }
+}
+
+// Starts a 'cdrdao' for blanking a CD.
+// Return: 0: OK, process succesfully launched
+//         1: error occured
+int CdDevice::blank(Project* parent, int fast, int speed, int eject,
+                    int reload)
+{
+  char *args[20];
+  int n = 0;
+  char devname[30];
+  char drivername[50];
+  char speedbuf[20];
+  char *execName;
+  const char *s;
+  int remoteFdArgNum = 0;
+
+  if ((status_ != DEV_READY && status_ != DEV_FAULT && status_ != DEV_UNKNOWN)
+      || process_ != NULL)
+    return 1;
+
+  if ((s = gnome_config_get_string(SET_CDRDAO_PATH)) != NULL)
+    execName = strdupCC(s);
+  else
+    execName = strdupCC("cdrdao");
+
+  args[n++] = execName;
+
+  args[n++] = "blank";
+
+  args[n++] = "--remote";
+
+  remoteFdArgNum = n;
+  args[n++] = NULL;
+
+  args[n++] = "-v0";
+
+  args[n++] = "--blank-mode";
+
+  if (fast)
+    args[n++] = "minimal";
+  else
+    args[n++] = "full";
+
+  if (speed > 0) {
+    sprintf(speedbuf, "%d", speed);
+    args[n++] = "--speed";
+    args[n++] = speedbuf;
+  }
+
+  if (eject)
+    args[n++] = "--eject";
+
+  if (reload)
+    args[n++] = "--reload";
+
+  args[n++] = "--device";
+  args[n++] = (char*)dev_.c_str();
+
+  if (driverId_ > 0) {
+    sprintf(drivername, "%s:0x%lx", driverName(driverId_), options_);
+    args[n++] = "--driver";
+    args[n++] = drivername;
+  }
+
+  args[n++] = NULL;
+  
+  assert(n <= 20);
+  
+  if (parent)
+    PROGRESS_POOL->start(*parent, this, _("Blanking CDRW"), false, false);
+  else
+    PROGRESS_POOL->start(this, _("Blanking CDRW"), false, false);
+
+  // Remove the SCSI interface of this device to avoid problems with double
+  // usage of device nodes.
+  delete scsiIf_;
+  scsiIf_ = NULL;
+
+  process_ = PROCESS_MONITOR->start(execName, args, remoteFdArgNum);
+
+  delete[] execName;
+
+  if (process_ != NULL) {
+    status_ = DEV_BLANKING;
+    action_ = A_BLANK;
+
+    if (process_->commFd() >= 0) {
+        Glib::signal_io().connect(bind(slot(*this, &CdDevice::updateProgress),
+                                       process_->commFd()),
+                                  process_->commFd(),
+                                  Glib::IO_IN | Glib::IO_HUP);
+    }
+    return 0;
+  }
+  else {
+    return 1;
+  }
+}
+
+void CdDevice::abortBlank()
+{
+  if (process_ != NULL && !process_->exited()) {
+    PROCESS_MONITOR->stop(process_);
+  }
 }
 
 void CdDevice::createScsiIf()
@@ -573,14 +861,7 @@ void CdDevice::createScsiIf()
     return;
 
   delete scsiIf_;
-
-  if (specialDevice_ != NULL && *specialDevice_ != 0) {
-    scsiIf_ = new ScsiIf(specialDevice_);
-  }
-  else {
-    sprintf(buf, "%d,%d,%d", bus_, id_, lun_);
-    scsiIf_ = new ScsiIf(buf);
-  }
+  scsiIf_ = new ScsiIf(dev_.c_str());
 
   if (scsiIf_->init() != 0) {
     delete scsiIf_;
@@ -627,6 +908,12 @@ const char *CdDevice::status2string(Status s)
   case DEV_RECORDING:
     ret = "Recording";
     break;
+  case DEV_READING:
+    ret = "Reading";
+    break;
+  case DEV_WAITING:
+    ret = "Waiting";
+    break;
   case DEV_BLANKING:
     ret = "Blanking";
     break;
@@ -668,24 +955,85 @@ const char *CdDevice::deviceType2string(DeviceType t)
   return ret;
 }
 
-CdDevice *CdDevice::add(int bus, int id, int lun, const char *vendor,
-			const char *product)
+
+/* reads configured devices from gnome settings
+ */
+void CdDevice::importSettings()
+{
+  int i, n;
+  char *s;
+  char buf[20];
+  CdDevice *dev;
+
+  n = gnome_config_get_int(SET_DEVICES_NUM);
+
+  if (n > 0) {
+    gnome_config_push_prefix(SET_SECTION_DEVICES);
+    
+    for (i = 0; i < n; i++) {
+      sprintf(buf, "%d", i);
+      s = gnome_config_get_string(buf);
+
+      if (s != NULL) {
+	if ((dev = CdDevice::add(s)) != NULL)
+	  dev->manuallyConfigured(true);
+      }
+    }
+
+    gnome_config_pop_prefix();
+  }
+}
+
+
+/* saves manually configured devices as gnome settings
+ */
+void CdDevice::exportSettings()
+{
+  int i, n;
+  char *s;
+  char buf[20];
+  CdDevice *drun;
+
+  gnome_config_clean_section(SET_SECTION_DEVICES);
+
+  for (drun = first(), n = 0; drun != NULL; drun = next(drun)) {
+    if (drun->manuallyConfigured()) {
+      n++;
+    }
+  }
+
+  if (n > 0) {
+    gnome_config_set_int(SET_DEVICES_NUM, n);
+
+    gnome_config_push_prefix(SET_SECTION_DEVICES);
+
+    for (drun = first(), i = 0; drun != NULL; drun = next(drun)) {
+      if (drun->manuallyConfigured()) {
+	sprintf(buf, "%d", i);
+	s = drun->settingString();
+	gnome_config_set_string(buf, s);
+	delete[] s;
+
+	i++;
+      }
+    }
+
+    gnome_config_pop_prefix();
+  }
+}
+
+CdDevice *CdDevice::add(const char* dev, const char *vendor,
+                        const char *product)
 {
   CdDevice *run, *pred, *ent;
 
   for (pred = NULL, run = DEVICE_LIST_; run != NULL;
        pred = run, run = run->next_) {
-    if (bus < run->bus() ||
-	(bus == run->bus() && id < run->id()) ||
-	(bus == run->bus() && id == run->id() && lun < run->lun())) {
-      break;
-    }
-    else if (bus == run->bus() && id == run->id() && lun == run->lun()) {
+    if (strcmp(run->dev(), dev) == 0)
       return run;
-    }
   }
 
-  ent = new CdDevice(bus, id, lun, vendor, product);
+  ent = new CdDevice(dev, vendor, product);
 
   if (pred != NULL) {
     ent->next_ = pred->next_;
@@ -751,29 +1099,24 @@ static char *nextToken(char *&p)
 static CdDevice *addImpl(char *s)
 {
   char *p;
-  int bus, id, lun;
   int driverId;
-  string vendor;
-  string model;
-  string device;
+  std::string dev;
+  std::string vendor;
+  std::string model;
+  std::string device;
   unsigned long options;
   char *val;
   CdDevice::DeviceType type;
-  CdDevice *dev;
+  CdDevice *cddev;
+
+  if (s[0] != '\'')
+    return NULL;
 
   p = s;
 
   if ((val = nextToken(p)) == NULL)
     return NULL;
-  bus = atoi(val);
-
-  if ((val = nextToken(p)) == NULL)
-    return NULL;
-  id = atoi(val);
-
-  if ((val = nextToken(p)) == NULL)
-    return NULL;
-  lun = atoi(val);
+  dev = val;
 
   if ((val = nextToken(p)) == NULL)
     return NULL;
@@ -803,18 +1146,13 @@ static CdDevice *addImpl(char *s)
     return NULL;
   options = strtoul(val, NULL, 0);
 
-  if ((val = nextToken(p)) == NULL)
-    return NULL;
-  device = val;
+  cddev = CdDevice::add(dev.c_str(), vendor.c_str(), model.c_str());
 
-  dev = CdDevice::add(bus, id, lun, vendor.c_str(), model.c_str());
-
-  dev->driverId(driverId);
-  dev->deviceType(type);
-  dev->driverOptions(options);
-  dev->specialDevice(device.c_str());
+  cddev->driverId(driverId);
+  cddev->deviceType(type);
+  cddev->driverOptions(options);
   
-  return dev;
+  return cddev;
 }
 
 CdDevice *CdDevice::add(const char *setting)
@@ -830,14 +1168,13 @@ CdDevice *CdDevice::add(const char *setting)
 
 
 
-CdDevice *CdDevice::find(int bus, int id, int lun)
+CdDevice *CdDevice::find(const char* dev)
 {
   CdDevice *run;
 
   for (run = DEVICE_LIST_; run != NULL; run = run->next_) {
-    if (bus == run->bus() && id == run->id() && lun == run->lun()) {
+    if (strcmp(run->dev(), dev) == 0)
       return run;
-    }
   }
 
   return NULL;
@@ -848,24 +1185,31 @@ void CdDevice::scan()
   int i, len;
   ScsiIf::ScanData *sdata = ScsiIf::scan(&len);
 
-  if (sdata == NULL)
-    return;
+  if (sdata) {
+    for (i = 0; i < len; i++)
+      CdDevice::add(sdata[i].dev.c_str(), sdata[i].vendor, sdata[i].product);
+    delete[] sdata;
+  }
 
-  for (i = 0; i < len; i++)
-    CdDevice::add(sdata[i].bus, sdata[i].id, sdata[i].lun, sdata[i].vendor,
-		  sdata[i].product);
-
-  delete[] sdata;
+#ifdef SCSI_ATAPI
+  sdata = ScsiIf::scan(&len, "ATAPI");
+  if (sdata) {
+    for (i = 0; i < len; i++)
+      CdDevice::add(sdata[i].dev.c_str(), sdata[i].vendor, sdata[i].product);
+    delete[] sdata;
+  }
+#endif
 }
 
-void CdDevice::remove(int bus, int id, int lun)
+void CdDevice::remove(const char* dev)
 {
   CdDevice *run, *pred;
 
   for (pred = NULL, run = DEVICE_LIST_; run != NULL;
        pred = run, run = run->next_) {
-    if (bus == run->bus() && id == run->id() && lun == run->lun()) {
-      if (run->status() == DEV_RECORDING || run->status() == DEV_BLANKING)
+    if (strcmp(run->dev(), dev) == 0) {
+      if (run->status() == DEV_RECORDING || run->status() == DEV_BLANKING ||
+	  run->status() == DEV_READING || run->status() == DEV_WAITING)
 	return;
 	
       if (pred != NULL)
@@ -931,46 +1275,3 @@ int CdDevice::updateDeviceStatus()
 
   return newStatus;
 }
-
-int CdDevice::updateDeviceProgress()
-{
-  int progress = 0;
-  int fd;
-  fd_set fds;
-  int maxFd = -1;
-  CdDevice *run;
-  struct timeval timeout = { 0, 0 };
-
-  blockProcessMonitorSignals();
-
-  FD_ZERO(&fds);
-
-  // collect set of file descriptors for 'select'
-  for (run = DEVICE_LIST_; run != NULL; run = run->next_) {
-    if (run->process_ != NULL && (fd = run->process_->commFd()) >= 0) {
-      FD_SET(fd, &fds);
-      if (fd > maxFd)
-	maxFd = fd;
-    }
-  }
-
-  if (maxFd < 0) {
-    unblockProcessMonitorSignals();
-    return 0;
-  }
-
-  if (select(maxFd + 1, &fds, NULL, NULL, &timeout) > 0) {
-    for (run = DEVICE_LIST_; run != NULL; run = run->next_) {
-      if (run->process_ != NULL && (fd = run->process_->commFd()) >= 0) {
-	if (FD_ISSET(fd, &fds)) {
-	  if (run->updateProgress())
-	    progress = 1;
-	}
-      }
-    }
-  }
-
-  unblockProcessMonitorSignals();
-
-  return progress;
-}
